from miniMinio.FileSystem import MinioFS, LocalFS


class Minio:
    def __init__(self, type):
        self.type = type
        pass

    def config(self, **kwargs):
        if self.type.lower() == 'minio':
            return MinioFS(**kwargs)
        elif self.type.lower() == 'local':
            return LocalFS(**kwargs)
        else:
            raise ValueError("Only Minio or Local Filesystem type allowed.")


import datetime
import os
import shutil
import sys
import time
from io import StringIO

import pandas as pd
import xlrd
from loguru import logger
from minio import Minio, ResponseError

from functions.utils import get_attributes_minio, get_attributes_local, get_attributes_single_local

exclude = {".minio.sys"}


def create_connection(hostname,
                      access_key=None,
                      secret_key=None,
                      secure=True):
    """

    :param hostname: string 'hostname:port' for Minio File server OR 'local' for local file server
    :param access_key: access key
    :param secret_key:
    :param secure:
    :return:
    """
    if hostname == "local" or hostname == "remote":
        return hostname
    else:
        client = Minio(hostname,
                       access_key=access_key,
                       secret_key=secret_key,
                       secure=secure)

    return client


def get_buckets(client, remote_dir=None):
    """
    :param client: Minio connection object, or 'local' for local filesystem. If remote, add in directory
    :return:
    """

    store = []
    if client == "local" or client == "remote":
        if client is "local":
            remote_dir = os.getcwd()
        for dirpath, dirs, filenames in os.walk(remote_dir, topdown=True):
            dirs[:] = set(dirs) - exclude
            for f in filenames:
                if f[0] == ".":
                    continue
                else:
                    path = os.path.abspath(os.path.join(dirpath, f))
                    store.append((path, datetime.datetime.strptime(time.ctime(os.path.getctime(path)),
                                                                   "%a %b %d %H:%M:%S %Y")))
    else:
        all_buckets = client.list_buckets()
        for bucket in all_buckets:
            store.append((bucket.name, bucket.creation_date))
    return store


def get_all_object_details_in_bucket(client, bucket_name=None, attributes=["object_name", "last_modified"],
                                     filter_object=("file", "folder")):
    """
    :param attributes: list Desired attributes to return for every object
    :param client: Minio connection object
    :param bucket_name: str bucket name, or file path of root to start search from
    :param filter_object: str 'file', 'folder', 'excel'
    :return:
    """
    store = []
    if client == "local" or client == "remote":
        if client is "local":
            bucket_name = os.getcwd()
        for path, subdirs, files in os.walk(bucket_name, topdown=True):
            subdirs[:] = set(subdirs) - exclude
            if ("file" in filter_object) or ("excel" in filter_object):
                # file or excel
                store = get_attributes_local(files, path, store, filter_object)
            if "folder" in filter_object:
                store = get_attributes_local(subdirs, path, store, filter_object)
    else:
        all_objects = client.list_objects(bucket_name, prefix='', recursive=True)
        for s3_object in all_objects:
            # Check if folder
            if (s3_object.object_name[-1] == "/") and ("folder" in filter_object):
                store = get_attributes_minio(attributes, s3_object, store)
                continue
            # Check if file
            elif (s3_object.object_name[-1] != "/") and ("file" in filter_object):
                store = get_attributes_minio(attributes, s3_object, store)
                continue
            # Check if excel
            elif "excel" in filter_object:
                if (s3_object.object_name[-4:] == "xlsx") or (s3_object.object_name[-3:] == "xls"):
                    store = get_attributes_minio(attributes, s3_object, store)
                    continue
    return store


def get_object_details_in_bucket(client, object_name, bucket_name=None, attributes=["object_name", "last_modified"]):
    """

    :param client:
    :param object_name:
    :param bucket_name:
    :param attributes:
    :return:
    """
    store = []
    if client == "local" or client == "remote":
        path = object_name
        store = get_attributes_single_local(path, attributes)

    else:
        all_objects = client.list_objects(bucket_name, prefix='', recursive=True)
        for s3_object in all_objects:
            # Check if folder
            if object_name.lower() not in s3_object.object_name.lower():
                continue
            else:
                store = get_attributes_minio(attributes, s3_object, store)

    return store[0]


def get_single_object_path(client, bucket_name, filename):
    # TODO: Test
    # TODO: local
    store = []
    store_error = []
    counter = [0, 0]
    output = []
    all_objects = client.list_objects(bucket_name, prefix='', recursive=True)
    for s3_object in all_objects:
        # Check if folder
        if filename.lower() not in s3_object.object_name.lower():
            counter[1] += 1
            store_error.append(getattr(s3_object, 'object_name'))
        else:
            counter[0] += 1
            store.append(getattr(s3_object, 'object_name'))
    output.extend([store, store_error, counter])
    return output


def get_single_object_from_bucket_to_file(client, object_name, dir_path, bucket_name=None):
    """
    :param client:
    :param bucket_name:
    :param object_name:
    :param dir_path: desired local path relative to the working directory
    :return: transfers files to dir_path
    """
    if client == "local" or client == "remote":
        shutil.copy2(object_name, dir_path)
    else:
        try:
            client.fget_object(bucket_name, object_name, dir_path)
        except ResponseError as err:
            logger.error(err)
    return


def get_single_object_content_bytes(client, object_name, bucket_name=None):
    """
    :param client:
    :param bucket_name:
    :param object_name:
    :return:
    """
    if client == "local" or client == "remote":
        in_file = open(object_name, "rb")
        data = in_file.read()
        return data
    else:
        result = []
        try:
            data = client.get_object(bucket_name, object_name)
            for d in data.stream(32 * 1024):
                result.append(d)
            return b''.join(result)
        except ResponseError as err:
            return err


def get_single_excel_file_content(client, excel_file_object, bucket_name=None):
    """
    :param client:
    :param bucket_name:
    :param excel_file_object:
    :return:
    """
    if client == "local" or client == "remote":
        return xlrd.open_workbook(excel_file_object)
    else:
        result = []
        try:
            data = client.get_object(bucket_name, excel_file_object)
            for d in data.stream(32 * 1024):
                result.append(d)
            return xlrd.open_workbook(file_contents=b''.join(result))
        except ResponseError as err:
            return err


def get_single_csv_file_content(client, csv_file_object, bucket_name=None):
    """
    Returns a dataframe from a csv

    :param client:
    :param csv_file_object:
    :param bucket_name:
    :return:
    """
    if client == "local" or client == "remote":
        return pd.read_csv(csv_file_object)
    else:
        result = []
        try:
            data = client.get_object(bucket_name, csv_file_object)
            for d in data.stream(32*1024):
                result.append(d)
            return pd.read_csv(StringIO(str(b''.join(result), 'utf-8')))
        except ResponseError as err:
            return err


def upload_dataframe_to_csv(client, csv_file_object_name, df, bucket_name=None):
    if client == "local" or client == "remote":
        df.to_csv(csv_file_object_name, index=False)
    # save object locally
    else:
        temp_csv_file = os.path.join(os.path.dirname(__file__), "to_write.csv")
        df.to_csv(temp_csv_file, index=False)
        try:
            print(client.fput_object(bucket_name, csv_file_object_name, temp_csv_file, content_type="application/csv"))
            os.remove(temp_csv_file)
        except ResponseError as err:
            print(err)
        return



from datetime import datetime
from unittest.mock import MagicMock
import pandas as pd
import pytest

from unittest import mock

import miniMinio


@pytest.fixture
@mock.patch('miniMinio.FileSystem.Minio')
def MinioFileSystem(Minio):
    fs = miniMinio.Minio(type='minio').config(**{
        "hostname": "test",
        "access_key": "test",
        "secret_key": "test",
        "secure": False
    })
    return fs


@pytest.fixture
def LocalFileSystem():
    return miniMinio.Minio(type='local').config()


def test_minio_lb(MinioFileSystem):
    """
    list buckets should be called once and return a list of tuples (name, creation_date)
    """
    bucket = MagicMock()
    bucket.name = 'test'
    bucket.creation_date = '2020-01-01'
    MinioFileSystem.__client__.list_buckets.return_value.__iter__.return_value = [bucket]
    output = MinioFileSystem.lb()
    MinioFileSystem.__client__.list_buckets.assert_called_once()
    assert output == [bucket.name]


def test_minio_ls(MinioFileSystem):
    object = MagicMock()
    object.object_name = 'test'
    MinioFileSystem.__client__.list_objects.return_value.__iter__.return_value = [object]
    output = MinioFileSystem.ls('bucket/filepath/test.txt')
    MinioFileSystem.__client__.list_objects.assert_called_once_with('bucket',
                                                                    prefix='filepath/test.txt',
                                                                    recursive=False)
    assert output == [object.object_name]


def test_minio_tree(MinioFileSystem):
    with pytest.raises(NotImplementedError):
        MinioFileSystem.tree('test')


def test_minio_stat(MinioFileSystem):
    object = MagicMock()
    object.object_name = 'test'
    object.is_dir = True
    object.last_modified = datetime(2020, 1, 1, 0, 0, 0)
    object.size = 1234
    MinioFileSystem.__client__.list_objects.return_value.__iter__.return_value = [object]
    output = MinioFileSystem.stat('bucket/testpath')
    MinioFileSystem.__client__.list_objects.assert_called_once_with('bucket', prefix='testpath', recursive=False)

    assert output == [{
        "name": "test",
        "type": "folder",
        "modified": "2020-01-01 00:00:00",
        "size": 1234
    }]


def test_minio_find(MinioFileSystem):
    object = MagicMock()
    object.object_name = 'test/filename.csv'
    MinioFileSystem.__client__.list_objects.return_value.__iter__.return_value = [object]
    output = MinioFileSystem.find('bucket/test/', 'filename.csv')
    MinioFileSystem.__client__.list_objects.assert_called_once_with('bucket', prefix='test/', recursive=True)
    assert output == ["test/filename.csv"]


@mock.patch('miniMinio.FileSystem.MinioFS._stream_file')
def test_minio_read_bytes(stream_file, MinioFileSystem):
    stream_file.return_value = [b'some', b'data', b'to', b'test']
    output = MinioFileSystem.read_bytes("bucket/test.txt")
    stream_file.assert_called_once_with("bucket/test.txt")
    assert output == b'somedatatotest'


@mock.patch('miniMinio.FileSystem.xlrd.open_workbook')
@mock.patch('miniMinio.FileSystem.MinioFS._stream_file')
def test_minio_read_excel(stream_file, open_workbook, MinioFileSystem):
    stream_file.return_value = [b'some', b'data', b'to', b'test']
    MinioFileSystem.read_excel('bucket/somefile.xls')
    stream_file.assert_called_once_with("bucket/somefile.xls")
    open_workbook.assert_called_once_with(file_contents=b'somedatatotest')


def test_minio_process_path(MinioFileSystem):
    bucket, prefix = MinioFileSystem._process_path('bucket/the/path/to/some/file.txt')
    assert bucket == 'bucket'
    assert prefix == 'the/path/to/some/file.txt'


def test_minio_stream_file(MinioFileSystem):
    file = MagicMock()
    file.stream.return_value = [b'some', b'bytes']
    MinioFileSystem.__client__.get_object.return_value=file

    output = MinioFileSystem._stream_file("bucket/file.txt")

    MinioFileSystem.__client__.get_object.assert_called_once_with(bucket_name="bucket", object_name="file.txt")
    assert output == [b'some', b'bytes']


def test_minio_rm(MinioFileSystem):
    MinioFileSystem.rm(["bucket/file1.txt", "bucket/file2.txt"])
    assert MinioFileSystem.__client__.remove_object.call_count == 2


@mock.patch('miniMinio.FileSystem.LocalFS.ls')
def test_local_lb(ls, LocalFileSystem):
    LocalFileSystem.lb()
    ls.assert_called_once_with(path='.', grep=None)


@mock.patch('miniMinio.FileSystem.os.listdir')
def test_local_ls(listdir, LocalFileSystem):
    LocalFileSystem.ls(path='.')
    listdir.assert_called_once_with('.')


@mock.patch('miniMinio.FileSystem.Path.iterdir')
def test_local_stat_dir(iterdir, LocalFileSystem):
    stats = MagicMock()
    stats.st_mtime = 123456789
    stats.st_size = 123

    file = MagicMock()
    file.name = 'file.txt'
    file.stat.return_value = stats
    file.is_dir.return_value = False
    iterdir.return_value = [file]
    attributes = LocalFileSystem.stat('.')
    iterdir.assert_called_once()
    assert attributes == [{"name": "file.txt",
                          "type": "file",
                          "modified": '1973-11-30 05:03:09',
                          "size": 123}]


@mock.patch('miniMinio.FileSystem.Path')
def test_local_stat_file(Path, LocalFileSystem):
    stat = MagicMock()
    stat.st_mtime = 123456789
    stat.st_size = 123

    file = MagicMock()
    file.stat.return_value = stat
    file.name = 'text.txt'
    Path.return_value = file
    Path('./text.txt').is_dir.return_value = False

    attributes = LocalFileSystem.stat('./text.txt')
    assert attributes == [{"name": "text.txt",
                           "type": "file",
                           "modified": '1973-11-30 05:03:09',
                           "size": 123}]


@mock.patch('miniMinio.FileSystem.os.walk')
def test_local_find(walk, LocalFileSystem):
    walk.return_value = [('.', ('dir',), ('file.txt',))]
    paths = LocalFileSystem.find('.', 'file.txt')
    assert paths == ["./file.txt"]


@mock.patch('os.remove')
def test_local_rm(remove, LocalFileSystem):
    LocalFileSystem.rm(['./file1.txt', './file2.txt'])
    assert remove.call_count == 2


@pytest.mark.parametrize('destination', ['s3://bucket/text_destination.txt', './text_destination.txt'])
@pytest.mark.parametrize('source', ['s3://bucket/text_source.txt', './text_source.txt'])
@mock.patch('miniMinio.FileSystem.shutil.copyfile')
@mock.patch('miniMinio.FileSystem.CopySource')
def test_cp(copysource, copyfile, MinioFileSystem, destination, source):
    MinioFileSystem.cp(source=source, destination=destination)
    if source == 's3://text_source.txt' and destination == './text_destination.txt':
        MinioFileSystem.__client.__fget_object.assert_called_once_with('bucket',
                                                                       object_name='bucket/text_source.txt',
                                                                       file_path='./text_destination.txt')
        MinioFileSystem.__client__.fput_object.assert_not_called()
        MinioFileSystem.__cluent__.copy_object.assert_not_called()
        copyfile.assert_not_called()

    if source == './text_source.txt' and destination == 's3://bucket/text_destination.txt':
        MinioFileSystem.__client__.fput_object.assert_called_once_with('bucket',
                                                                       object_name='text_destination.txt',
                                                                       file_path='./text_source.txt')
        MinioFileSystem.__client__.fget_object.assert_not_called()
        MinioFileSystem.__client__.copy_object.assert_not_called()
        copyfile.assert_not_called()

    if source == './text_source.txt' and destination == './text_destination.txt':
        copyfile.assert_called_once_with(source, destination)
        MinioFileSystem.__client__.fget_object.assert_not_called()
        MinioFileSystem.__client__.copy_object.assert_not_called()
        MinioFileSystem.__client__.fput_object.assert_not_called()

    if source == 's3://bucket/text_source.txt' and destination == 's3://bucket/text_destination.txt':
        MinioFileSystem.__client__.copy_object.assert_called_once_with('bucket', 'text_destination.txt', copysource('bucket', 'text_source.txt'))
        MinioFileSystem.__client__.fget_object.assert_not_called()
        copyfile.assert_not_called()
        MinioFileSystem.__client__.fput_object.assert_not_called()


@mock.patch('miniMinio.FileSystem.BytesIO')
def test_minio_write_csv(bytesio, MinioFileSystem):
    df = MagicMock()
    csv_data = MagicMock()
    csv_bytes = [1, 2, 3, 4]
    csv_data.encode.return_value = csv_bytes
    df.to_csv.return_value = csv_data
    data = [1, 2, 3]
    bytesio.return_value = data
    MinioFileSystem.write_csv(df, 'bucket/folder/df.csv')
    MinioFileSystem.__client__.put_object.assert_called_once_with('bucket', 'folder/df.csv', data=data,
                                                                   length=len(csv_bytes), content_type='application/csv')


def test_local_write_csv(LocalFileSystem):
    df = MagicMock()
    LocalFileSystem.write_csv(df, 'bucket/folder/df.csv')
    df.to_csv.assert_called_once_with('bucket/folder/df.csv', index=None, header=None)



# Abstract Factory
import datetime
import os
import time
from abc import ABC, abstractmethod

from minio import Minio

exclude = {".minio.sys"}


class ConnectionClient():
    def set_file_server_type(self, file_server):
        if file_server == "local":
            return LocalFactory()
        if file_server == "minio":
            return MinioFactory()
        raise TypeError("Unknown file server type")


class Product(ABC):
    @abstractmethod
    def create_connection(self, hostname, access_key, secret_key, secure):
        pass

    @abstractmethod
    def get_buckets(self, remote_dir):
        pass

    @abstractmethod
    def get_all_object_details_in_bucket(self, bucket_name=None, attributes=["object_name", "last_modified"],
                                         filter_object=("file", "folder")):
        pass

    @abstractmethod
    def get_object_details_in_bucket(self, object_name, bucket_name=None, attributes=["object_name", "last_modified"]):
        pass

    @abstractmethod
    def get_object_path(self, bucket_name, file_name):
        pass

    @abstractmethod
    def get_single_object_from_bucket_to_file(self, object_name, dir_path, bucket_name=None):
        pass

    @abstractmethod
    def get_single_object_content_bytes(self, object_name, bucket_name=None):
        pass

    @abstractmethod
    def get_single_excel_file_content(self, excel_file_object, bucket_name=None):
        pass

    @abstractmethod
    def get_single_csv_file_content(self, csv_file_object, bucket_name=None):
        pass

    @abstractmethod
    def upload_dataframe_to_csv(self, csv_file_object_name, df, bucket_name=None):
        pass


class LocalFactory(Product):
    def create_connection(self, hostname, access_key, secret_key, secure):
        return "local"

    def get_buckets(self, remote_dir):
        store = []
        remote_dir = os.getcwd()
        for dirpath, dirs, filenames in os.walk(remote_dir, topdown=True):
            dirs[:] = set(dirs) - exclude
            for f in filenames:
                if f[0] == ".":
                    continue
                else:
                    path = os.path.abspath(os.path.join(dirpath, f))
                    store.append((path, datetime.datetime.strptime(time.ctime(os.path.getctime(path)),
                                                                   "%a %b %d %H:%M:%S %Y")))


class MinioFactory(Product):
    def __init__(self):
        self.client = ''

    def create_connection(self, hostname, access_key, secret_key, secure):
        client = Minio(hostname, access_key, secret_key, secure)
        self.client = client

    def get_buckets(self, remote_dir):
        client = self.client
        store = []
        all_buckets = client.list_buckets()
        for bucket in all_buckets:
            store.append((bucket.name, bucket.creation_date))
        return store



import datetime
import os
import time


def get_attributes_minio(attributes, s3_object, store):
    attribute_list = []
    for attribute in attributes:
        attribute_list.append(getattr(s3_object, attribute))
    store.append(tuple(attribute_list))
    return store


def get_attributes_local(files, path, store, filter_object):
    for name in files:
        if "excel" in filter_object:
            if (name[-4:] == "xlsx") or (name[-3:] == "xls"):
                store.append((os.path.join(path, name),
                              datetime.datetime.strptime(time.ctime(os.path.getmtime(path)), "%a %b %d %H:%M:%S %Y")))
        else:
            store.append((os.path.join(path, name),
                          datetime.datetime.strptime(time.ctime(os.path.getmtime(path)), "%a %b %d %H:%M:%S %Y")))
    return store


def path_leaf_last(path):
    head, tail = os.path.split(path)
    head = os.path.basename(head)
    return tail or head


def get_attributes_single_local(path, store):
    store.append((path,
                  datetime.datetime.strptime(time.ctime(os.path.getmtime(path)), "%a %b %d %H:%M:%S %Y")))
    return store



import os
from unittest.mock import Mock

import pytest

from miniMinio import create_connection, get_buckets, get_all_object_details_in_bucket, \
    get_single_object_from_bucket_to_file, get_single_object_content_bytes
from miniMinio.utils import path_leaf_last


class MockBucket:
    def __init__(self, name, creation_date):
        self.name = name
        self.creation_date = creation_date


class MockObject:
    def __init__(self, object_name, creation_date):
        self.object_name = object_name
        self.creation_date = creation_date
        self.is_dir = False


@pytest.mark.parametrize("hostname",
                         [(("minio")),
                          (("local"))])
def test_create_connection(hostname, snapshot):
    access_key = "test_access_key"
    secret_key = "test_secret_key"
    client = create_connection(hostname, access_key, secret_key)
    if type(client) is str:
        snapshot.assert_match(client)
    # No connections actually made


def test_get_buckets(snapshot):
    mock_client = Mock()
    mock_client.list_buckets.return_value = [
        MockBucket("bucket1", "bucket1_date"),
        MockBucket("bucket2", "bucket2_date")
    ]
    result = get_buckets(mock_client)
    snapshot.assert_match(result)


@pytest.mark.parametrize("mock_client", [
    (("local")),
    (("remote"))
])
def test_get_buckets_local(mock_client, snapshot):
    if mock_client == "remote":
        mock_remote_dir = os.path.join(__file__, 'data')
    else:
        mock_remote_dir = None
    results = get_buckets(mock_client, mock_remote_dir)
    names = [result[0] for result in results]
    names = [path_leaf_last(name) for name in names]
    names.sort()
    snapshot.assert_match(names)


@pytest.mark.parametrize('filter_object, ext', [
    (("file"), (None)),
    (("folder"), (None)),
    (("file", "folder"), (None)),
    (('file'), ('xlsx', 'xls')),
    (('file', 'folder'), ('xlsx', 'xls'))
])
def test_get_all_object_details_in_bucket(filter_object, ext, snapshot):
    mock_client = Mock()
    mock_client.list_objects.return_value = [
        MockObject("bucket1/object1.pdf", "pdf_date"),
        MockObject("bucket2/folder/", "folder_date"),
        MockObject("bucket1/object2.xlsx", "xlsx_date"),
        MockObject("bucket2/object3.xls", "xls_date")
    ]
    for idx, mock_object in enumerate(mock_client.list_objects.return_value):
        if idx == 1:
            mock_object.is_dir = True
    bucket_name = "placeholder"
    attributes = ["creation_date"]
    filter_prefix = ''
    results = get_all_object_details_in_bucket(mock_client, bucket_name, attributes, filter_object, filter_prefix, ext)
    names = [result[0] for result in results]
    names = [path_leaf_last(name) for name in names]
    names.sort()
    snapshot.assert_match(names)


@pytest.mark.parametrize('filter_object, ext', [
    (("file"), (None)),
    (("folder"), (None)),
    (("file", "folder"), (None)),
    (('file'), ('xlsx', 'xls')),
    (('file', 'folder'), ('xlsx', 'xls'))
])
def test_get_all_object_details_in_bucket_local_fs(filter_object, ext, snapshot):
    bucket_name = None
    attributes = ["object_name"]
    filter_prefix = ''
    results = get_all_object_details_in_bucket('local', bucket_name, attributes, filter_object, filter_prefix, ext)
    names = [result[0] for result in results]
    names = [path_leaf_last(name) for name in names]
    names.sort()
    snapshot.assert_match(names)


@pytest.mark.parametrize('filter_object, ext', [
    (("file"), (None)),
    (("folder"), (None)),
    (("file", "folder"), (None)),
    (('file'), ('xlsx', 'xls')),
    (('file', 'folder'), ('xlsx', 'xls'))
])
def test_get_all_object_details_in_bucket_remote_fs(filter_object, ext, snapshot):
    bucket_name = os.path.join(os.path.dirname(__file__), 'data')
    attributes = ["object_name"]
    filter_prefix = ''
    results = get_all_object_details_in_bucket("remote", bucket_name, attributes, filter_object, filter_prefix, ext)
    names = [result[0] for result in results]
    names = [path_leaf_last(name) for name in names]
    names.sort()
    snapshot.assert_match(names)


def test_get_single_objects_from_bucket_to_file():
    environment = os.environ.get('ENVIRON', "local")
    if environment == "local":
        hostname = "localhost:9000"
    elif environment == "test":
        hostname = "minio:9000"
    else:
        raise EnvironmentError("No ENV Variable!")
    client = create_connection(hostname, "minioadmin", "minioadmin", secure=False)
    object_name = "file1.txt"
    dir_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "receive_file1.txt")
    bucket_name = "buckettest"
    get_single_object_from_bucket_to_file(client, object_name, dir_path, bucket_name)
    os.remove(dir_path)  # remove to see file
    return


def test_get_single_objects_from_bucket_to_file_fs():
    client = create_connection("remote", "minioadmin", "minioadmin", secure=False)
    object_name_list = os.path.join(os.path.realpath(os.path.dirname(__file__)), "data", "buckettest", "file1.txt")
    dir_path = os.path.join(os.path.realpath(os.path.dirname(__file__)), "receive_file1.txt")
    get_single_object_from_bucket_to_file(client, object_name_list, dir_path)
    os.remove(dir_path)  # remove to see file
    return


def test_get_single_object_content_bytes(snapshot):
    environment = os.environ.get('ENVIRON', "local")
    if environment == "local":
        hostname = "localhost:9000"
    elif environment == "test":
        hostname = "minio:9000"
    else:
        raise EnvironmentError("No ENV Variable!")
    client = create_connection(hostname, "minioadmin", "minioadmin", secure=False)
    bucket_name = "buckettest"
    object_name = "BigBook.xlsx"
    result = get_single_object_content_bytes(client, object_name, bucket_name)
    snapshot.assert_match(result)


def test_get_single_object_content_bytes_fs(snapshot):
    client = create_connection("remote", "minioadmin", "minioadmin", secure=False)
    object_name = os.path.join(os.path.realpath(os.path.dirname(__file__)), "data", "buckettest", "BigBook.xlsx")
    result = get_single_object_content_bytes(client, object_name)
    snapshot.assert_match(result)



import os
import re
import shutil
from abc import ABC, abstractmethod
from io import BytesIO

import pytz
import pandas as pd
import xlrd
from loguru import logger
from minio import Minio
from pathlib import Path
from datetime import datetime

# prefix components:
from minio.select import SelectRequest, CSVInputSerialization, CSVOutputSerialization
from minio.commonconfig import CopySource

space = '    '
branch = '│   '
# pointers:
tee = '├── '
last = '└── '


class FileSystem(ABC):
    def __init__(self, **kwargs):
        self.__client__ = kwargs.get("client", None)

    @abstractmethod
    def lb(self):
        pass

    @abstractmethod
    def ls(self, path, recursive, grep=None):
        pass

    @abstractmethod
    def tree(self, path: Path, prefix: str = ''):
        pass

    @abstractmethod
    def stat(self, path: str):
        pass

    @abstractmethod
    def find(self, path: str, filename: str):
        pass

    @abstractmethod
    def read_bytes(self, filename: str):
        pass

    @abstractmethod
    def read_excel(self, filename: str):
        pass

    @abstractmethod
    def rm(self, paths: str):
        pass

    @abstractmethod
    def write_csv(self, df: pd.DataFrame, filename: str):
        pass

    def grep(self, iterable: iter, regex: str) -> list:
        return [obj for obj in iterable if re.match(rf'{regex}', obj)]

    def cp(self, source, destination):
        """
        Copy a file from:
            1) Minio to local e.g cp("s3://bucket/folder/file.txt", "./file.txt")
            2) local to Minio e.g cp("./file.txt", "s3://bucket/folder/file.txt")
            3) local to local e.g cp("./file.txt", "./newfile/txt")
            4) Minio to minio e.g cp("s3://bucket/folder/file.txt", "s3://bucket/folder/newfile.txt")

        Prefix minio paths with a s3://
        """
        source_type = 'local'
        destination_type = 'local'
        if 's3://' in source:
            source_type = 'remote'
            source_path = source.split('s3://')[1]
            source_bucket, source_filename = self._process_path(source_path)
        if 's3://' in destination:
            destination_type = 'remote'
            destination_path = destination.split('s3://')[1]
            destination_bucket, destination_filename = self._process_path(destination_path)

        logger.info(f"Copying from {source_type}: {source} to {destination_type}: {destination}")

        if source_type == 'remote' and destination_type == 'local':
            self.__client__.fget_object(source_bucket, object_name=source_filename, file_path=destination)

        if source_type == 'local' and destination_type == 'remote':
            self.__client__.fput_object(destination_bucket, object_name=destination_filename, file_path=source)

        if source_type == 'local' and destination_type == 'local':
            shutil.copyfile(source, destination)

        if source_type == 'remote' and destination_type == 'remote':
            self.__client__.copy_object(destination_bucket, destination_filename,
                                        CopySource(source_bucket, source_filename))


class MinioFS(FileSystem):
    def __init__(self, hostname, access_key=None, secret_key=None, secure=True):
        client = Minio(hostname,
                       access_key=access_key,
                       secret_key=secret_key,
                       secure=secure)
        super().__init__(client=client)

    def lb(self, grep=None):
        """
        List all buckets.
        :return: list of tuples (name, creation date)
        """
        buckets = [bucket.name for bucket in self.__client__.list_buckets()]
        if grep:
            return self.grep(buckets, grep)
        else:
            return buckets

    def ls(self, path: str, recursive: bool = False, grep: str = None):
        """
        :param recursive:
        :param path: <bucket>/path/to/folder/in/bucket
        :param grep:
        :return:
        """
        bucket, prefix = self._process_path(path)
        objects = [obj.object_name for obj in self.__client__.list_objects(bucket,
                                                                           prefix=prefix,
                                                                           recursive=recursive)]
        if grep:
            return self.grep(objects, grep)
        else:
            return objects
        pass

    def tree(self, path: Path, prefix: str = ''):
        raise NotImplementedError("Tree list is not implemented for minio file system yet. Sorry ):")
        pass

    def stat(self, path):
        """

        :param path: path to directory or file. If directory, it will stat for all files in the directory.
        :return:
        """
        file_attributes = []
        bucket, prefix = self._process_path(path)
        for obj in self.__client__.list_objects(bucket, prefix=prefix, recursive=False):
            file_attributes.append({
                "name": obj.object_name,
                "type": "folder" if obj.is_dir else "file",
                "modified": datetime.strftime(obj.last_modified, "%Y-%m-%d %H:%M:%S"),
                "size": obj.size,
            })
        return file_attributes
        pass

    def find(self, path, filename):
        """
        Recursively look for a file starting from given path. Match exact, case sensitive.
        :param path: Starting directory to start looking top down
        :param filename: name of file to search for
        :return: path to file
        """
        filepaths = []
        bucket, prefix = self._process_path(path)
        for obj in self.__client__.list_objects(bucket, prefix=prefix, recursive=True):
            if filename in obj.object_name:
                filepaths.append(obj.object_name)

        return filepaths

    def read_bytes(self, filename):
        data = self._stream_file(filename)
        return b''.join(data)

    def read_excel(self, filename):
        """
        returns xlrd workbook object
        :param filename:
        :return: An instance of the :class:`~xlrd.book.Book` class.
        """
        data = self._stream_file(filename)
        return xlrd.open_workbook(file_contents=b''.join(data))

    def _process_path(self, path) -> (str, str):
        """
        Split path string into bucket and the prefix for usage in list_objects for the s3 api.
        :param path: Path in s3. Include bucket name in path.
        :return:
        """
        bucket = path.split('/')[0]
        prefix = '/'.join(path.split('/')[1:])
        return bucket, prefix

    def _stream_file(self, filename) -> list:
        data = []
        bucket, filepath = self._process_path(filename)
        file = self.__client__.get_object(bucket_name=bucket, object_name=filepath)
        for data_stream in file.stream(32 * 1024):
            data.append(data_stream)
        return data

    def query_csv(self, filename: str, sql: str):
        bucket, filepath = self._process_path(filename)
        data = self.__client__.select_object_content(bucket,
                                                     filepath,
                                                     SelectRequest(sql,
                                                                   CSVInputSerialization(file_header_info="USE"),
                                                                   CSVOutputSerialization(),
                                                                   request_progress=True)
                                                     )

        headers = self.__client__.select_object_content(bucket,
                                                        filepath,
                                                        SelectRequest("select * from s3object limit 1",
                                                                      CSVInputSerialization(file_header_info="NONE"),
                                                                      CSVOutputSerialization(),
                                                                      request_progress=True)
                                                        )
        bytearray = b""
        for line in headers.stream():
            bytearray = bytearray + line
        for line in data.stream():
            bytearray = bytearray + line
        return pd.read_csv(BytesIO(bytearray))

    def rm(self, paths):
        """
        :param paths:
        """
        if type(paths) != list:
            raise TypeError("Invalid type: paths is not a list.")
        filenames = set(map(lambda x: '/'.join(x.split('/')[1:]), paths))
        bucket_unique = set((map(lambda x: x.split('/')[0], paths)))
        if len(bucket_unique) != 1:
            raise ValueError(f"Removing files across multiple buckets ({bucket_unique}) is not permitted.")
        else:
            bucket = bucket_unique.pop()
        for filename in filenames:
            self.__client__.remove_object(bucket_name=bucket, object_name=filename)

    def write_csv(self, df, filename):
        bucket, filepath = self._process_path(filename)
        csv_bytes = df.to_csv(index=None, header=None).encode('utf-8')
        csv_buffer = BytesIO(csv_bytes)
        self.__client__.put_object(bucket, filepath, data=csv_buffer,
                                    length=len(csv_bytes), content_type='application/csv')



class LocalFS(FileSystem):
    def lb(self, grep=None):
        """
        Since local fs has no such thing as buckets, we will just do a ls on the pwd.
        :param grep: regex pattern that will be matched using re.match()
        """
        return self.ls(path='.', grep=grep)

    def ls(self, path, recursive=False, grep=None):
        """
        :param path: path to list directory
        :param grep: regex pattern that will be matched using re.match()
        :return: list of files & folders in directory
        """
        if recursive:
            logger.warn("Recusrive mode only supported for minio file system. Will continue to run ls in non recursive"
                        "mode.")

        items = os.listdir(path)
        if grep:
            return self.grep(items, grep)
        else:
            return items

    def tree(self, path: Path, prefix: str = ''):
        """A recursive generator, given a directory Path object
        will yield a visual tree structure line by line
        with each line prefixed by the same characters

        example: fs.tree(Path()) to get tree for current working dir.
       for line in fs.tree(Path()):
            print(line)
        """
        contents = list(path.iterdir())
        # contents each get pointers that are ├── with a final └── :
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            yield prefix + pointer + path.name
            if path.is_dir():  # extend the prefix and recurse:
                extension = branch if pointer == tee else space
                # i.e. space because last, └── , above so no more |
                yield from self.tree(path, prefix=prefix + extension)

    def stat(self, path):
        """
        Get file statistics for files in the directory
        :param path: path to directory
        :return: list of file statistics
        """
        file_attributes = []

        if Path(path).is_dir():
            for file in Path(path).iterdir():
                stats = file.stat()
                file_attributes.append({
                    "name": file.name,
                    "type": "folder" if file.is_dir() else "file",
                    "modified": datetime.strftime(
                        datetime.fromtimestamp(stats.st_mtime, pytz.timezone("Asia/Singapore")), "%Y-%m-%d %H:%M:%S"),
                    "size": stats.st_size,
                })
        else:
            file = Path(path)
            stats = file.stat()
            file_attributes.append({
                "name": file.name,
                "type": "file",
                "modified": datetime.strftime(datetime.fromtimestamp(stats.st_mtime, pytz.timezone("Asia/Singapore")),
                                              "%Y-%m-%d %H:%M:%S"),
                "size": stats.st_size,
            })

        return file_attributes

    def find(self, path, filename):
        filepaths = []
        for root, dirs, files in os.walk(path):
            if filename in files:
                filepaths.append(os.path.join(root, filename))
        return filepaths

    def read_bytes(self, filename):
        """
        Read bytes of a file.
        :param filename: path to file to read.
        :return:
        """
        file = open(filename, "rb")
        return file.read()

    def read_excel(self, filename):
        """

        :param filename:
        :return: An instance of the :class:`~xlrd.book.Book` class.
        """
        return xlrd.open_workbook(filename)

    def rm(self, paths) -> None:
        if type(paths) != list:
            raise TypeError("Invalid type: paths is not a list.")
        for path in paths:
            logger.info(f"Removing file: {path}")
            os.remove(path)
        return

    def write_csv(self, df, filename):
        df.to_csv(filename, index=None, header=None)

